"""Memory processing."""

import os
import logging

import torch
from sentence_transformers import SentenceTransformer, util

from ..utils import ConfigManager, LocalAssistantException

class MemoryExtension:
    """Control users' memory."""
    def __init__(self, model_name: str):
        self.config = ConfigManager()
        self.utils_ext = self.config.utils_ext

        try:
            temp_path: str = os.path.join\
                (self.utils_ext.model_path, 'Sentence_Transformer', model_name)
            self.model = SentenceTransformer(temp_path, local_files_only=True)
        except Exception as err:
            logging.error('Can not load model due to: %s', err)
            raise LocalAssistantException('Can not load model.') from err

    def encode_memory(self) -> None:
        """Encode user's memory."""
        self.config.get_config_file()

        data: list = []
        role: list = []

        logging.info('Get history\'s data.')
        temp_path: str = os.path.join\
            (self.utils_ext.user_path, self.config.data['users'], 'history')
        for history_file in os.scandir(temp_path):
            # if not json file, continue.
            if not history_file.is_file():
                continue
            if not history_file.name.endswith('.json'):
                continue

            # the rest is .json file.
            logging.debug('Reading %s.', history_file.name)

            json_data = self.utils_ext.read_json_file(history_file.path)

            for conversation in json_data:
                if conversation["role"] == "system": # skip system
                    continue
                role.append(conversation["role"])
                data.append(conversation["content"])

        # save memory to json file.
        logging.debug("Transfer data to .json")
        data_to_json: dict = {}
        for index, item in enumerate(data):
            data_to_json.update({
                index: {
                    "role": role[index],
                    "content": item,
                }
            })

        temp_path = os.path.join\
            (self.utils_ext.user_path, self.config.data['users'], 'memory', 'all_memory.json')
        self.utils_ext.write_json_file(temp_path, data_to_json)

        # encode
        encoded_data = self.model.encode(data, convert_to_tensor=True, show_progress_bar=True)

        temp_path = os.path.join\
            (self.utils_ext.user_path, self.config.data['users'], 'memory', 'encoded_memory.pt')
        torch.save(encoded_data, temp_path)

    def ask_query(self, question: str, top_k: int = 0) -> list[str]:
        """Ask query function."""
        self.config.get_config_file()

        if top_k == 0: # user didnt add top k when chat.
            top_k=int(self.config.data["top_k_memory"])

        temp_path: str = os.path.join\
            (self.utils_ext.user_path, self.config.data['users'], 'memory', 'encoded_memory.pt')
        logging.info('Loading memory from encoded data.')
        try:
            encoded_data = torch.load(temp_path, weights_only=True)
        except FileNotFoundError:
            logging.info('Encoded data not found. Encode new data.')
            self.encode_memory()
            encoded_data = torch.load(temp_path, weights_only=True)

        # encode query.
        encoded_question = self.model.encode(question, convert_to_tensor=True)

        # we only ask one question
        hits = util.semantic_search(encoded_question, encoded_data, top_k=top_k)[0]

        # get json file.
        data: dict = {}
        logging.debug('Get json file.')
        temp_path = os.path.join\
            (self.utils_ext.user_path, self.config.data['users'], 'memory', 'all_memory.json')
        data = self.utils_ext.read_json_file(temp_path)

        # Concluding.
        result: list = []
        for hit in hits:
            if hit['score'] < 0.5:
                break
            pointer: dict = data[str(hit["corpus_id"])]
            whose: str = 'Me: ' if pointer['role'] == 'user' else 'You: '

            result.append(f"{whose}'{pointer['content']}'")
            logging.debug("Use memory: %s'%s'", whose, pointer['content'])

        return result



"""Chat processing."""

import logging
import os
from threading import Thread

from transformers import AutoTokenizer, AutoModelForCausalLM,\
    TextIteratorStreamer, BitsAndBytesConfig

from ..utils import ConfigManager, LocalAssistantException
from .memory import MemoryExtension
from .docs import DocsQuestionAnswerExtension

class ChatExtension():
    """Chat extention for LocalAssistant."""
    def __init__(self):
        self.config = ConfigManager()
        self.utils_ext = self.config.utils_ext

    def _load_local_model(self, model_name: str) -> tuple:
        """
        Load text generation model with wanted bits.

        Args:
            model_name (str): Name of model.

        Raises:
            LocalAssistantException: invalid bits.

        Returns:
            tuple: (text generation, tokenizer)
        """
        path: str = os.path.join(self.utils_ext.model_path, 'Text_Generation', model_name)
        kwarg = dict(local_files_only=True,use_safetensors=True, device_map="auto",)

        used_bit = self.config.data["load_in_bits"]
        match used_bit:
            case '8':
                kwarg.update({'quantization_config': BitsAndBytesConfig(load_in_8bit=True)})
            case '4':
                kwarg.update({'quantization_config': BitsAndBytesConfig(load_in_4bit=True)})
            case 'None':
                pass
            case _:
                logging.error('Invalid bits! We found: %s.', used_bit)
                raise LocalAssistantException(f"Invalid bits! We found: {used_bit}.")
        kwarg.update()
        return (
            AutoModelForCausalLM.from_pretrained(path, **kwarg ),
            AutoTokenizer.from_pretrained(os.path.join(path, 'Tokenizer'), **kwarg)
        )

    @staticmethod
    def _chat(
            history: list,
            text_generation_model: AutoModelForCausalLM,
            tokenizer_model: AutoTokenizer,
            max_new_tokens: int,
            **kwargs,
        ) -> dict:
        """
        Simple chat system.

        Args:
            history (list): list of history.
            text_generation_model (AutoModelForCausalLM): text generation model.
            tokenizer_model (AutoTokenizer): tokenizer of text generation model.
            max_new_tokens (int): max tokens to generate.
            **kwargs: arguments for `apply_chat_template`. Used for memory, etc.

        Returns:
            dict: assistant's whole output.
        """
        # format history.
        format_history = tokenizer_model\
            .apply_chat_template(history, tokenize=False, add_generation_prompt=True, **kwargs)

        input_token = tokenizer_model(format_history, return_tensors="pt", add_special_tokens=False)

        # move token to device.
        input_token = {key: tensor.to(text_generation_model.device)\
            for key, tensor in input_token.items()}

        # make streamer.
        streamer = TextIteratorStreamer(tokenizer_model, skip_prompt=True)

        # threading the generation
        kwargs = dict(input_token, streamer=streamer, max_new_tokens=max_new_tokens, do_sample=True)
        thread = Thread(target=text_generation_model.generate, kwargs=kwargs)
        thread.start()

        full_output: str = ''
        for output in streamer:
            output = output.removesuffix('<|im_end|>')

            full_output += output
            print(output, end='', flush=True)

        return {"role": "assistant", "content": full_output}

    def chat_with_limited_lines(
            self,
            text_generation_model_name: str = '',
            lines: int = 1,
            max_new_tokens: int = 500,
        ):
        """
        Chat with models for limited lines. Recommend for fast chat as non-user. (no history saved)
        
        Args:
            text_generation_model_name (str): text generation model's name, use config if blank.
            lines (int): lines of chat (not count 'assistant'), default as 1.
            max_new_tokens (int): max tokens to generate, default as 500.
        """

        if lines < 1:
            raise LocalAssistantException("Argument 'lines' should not have non-positive value.")

        history: list = [
            {
                "role": "system", 
                "content": f"You are an Assistant named LocalAssistant (Locas). \
You only have {lines} lines, give the user the best supports as you can."
            },
        ]

        if text_generation_model_name == '':
            self.config.get_config_file()
            self.config.check_for_exist_model(1)
            text_generation_model_name = self.config.data['models']['Text_Generation']
            logging.info('No text generation model, use %s instead.', text_generation_model_name)

        # load model
        logging.debug('Begin to load models.')
        text_generation_model, tokenizer_model = self._load_local_model(text_generation_model_name)
        logging.debug('Done loading models.')

        # chat with limited lines
        print(f"\nStart chatting in {lines} line(s) with '{text_generation_model_name}'\
for text generation.\n\nType 'exit' to exit.", end='')

        for _ in range(lines):
            prompt: str = input('\n\n>> ')
            print()

            if prompt.lower() in ('exit', 'exit()'):
                break

            # append chat to history.
            history.append({"role": "user", "content": prompt,})

            reply = self._chat(history, text_generation_model, tokenizer_model, max_new_tokens)
            history.append(reply)

        # If user want to continue. Sometimes the conversation is cool I guess...
        while True:
            # If don't want to, end this loop
            print("\n\n------------------------------------")
            if input(f"Finished {lines} lines. Want to keep chatting? [y/n]: ").lower() != 'y':
                print("------------------------------------")
                break
            print("------------------------------------", end='')

            prompt: str = input('\n\n>> ')
            print()

            if prompt.lower() in ('exit', 'exit()'):
                break

            # append chat to history.
            history.append({"role": "user", "content": prompt,})

            reply = self._chat(history, text_generation_model, tokenizer_model, max_new_tokens)
            history.append(reply)

    def chat_with_history(
            self,
            text_generation_model_name: str = '',
            user: str = 'default',
            max_new_tokens: int = 500,
            memory_enable: bool = False,
            sentence_transformer_model_name: str = '',
            top_k_memory: int = 0,
            encode_at_start: bool = False,
        ):
        """
        Chat with models with unlimited lines. History will be saved.
        
        Args:
            text_generation_model_name (str): text generation model's name, use config if blank.
            user (str): chat by user, default as 'default'.
            max_new_tokens (int): max tokens to generate, default as 500.
            
            memory_enable (bool): enable memory function, default as False.
            sentence_transformer_model_name (str): sentence transformer model's name, \
                use config if blank.
            top_k_memory (int): how much memory you want to recall.
            encode_at_start (bool): encode memory before chating.
        """

        self.config.get_config_file()
        self.config.check_for_exist_user(user)

        if text_generation_model_name == '':
            self.config.check_for_exist_model(1)
            text_generation_model_name = self.config.data['models']['Text_Generation']
            logging.info('No text generation model, use %s instead.', text_generation_model_name)

        if memory_enable and sentence_transformer_model_name == '':
            self.config.check_for_exist_model(2)
            sentence_transformer_model_name = self.config.data['models']['Sentence_Transformer']
            logging.info('No sentence transformer model, \
use %s instead.', sentence_transformer_model_name)

        # load model
        logging.debug('Begin to load models.')
        text_generation_model, tokenizer_model = self._load_local_model(text_generation_model_name)
        if memory_enable:
            memory_ext = MemoryExtension(sentence_transformer_model_name) # --- start here ---
            if encode_at_start:
                logging.info('Encoding at start.')
                memory_ext.encode_memory()
        logging.debug('Done loading models.')

        # load chat history.
        logging.debug('Loading history.')
        chat_history, chat_name = self.utils_ext.load_chat_history(user)

        # user typed 'exit'
        if chat_name == '':
            return

        # chat with history.
        line_print: str = f"Start chatting as user '{user}' with \
'{chat_name}' for history, '{text_generation_model_name}' for text generation"
        if memory_enable:
            line_print += f", '{sentence_transformer_model_name}' for sentence transformer"
        print(f"{line_print}.\n\nType 'exit' to exit.", end='')

        while True:
            prompt: str = input('\n\n>> ')
            print()

            if prompt.lower() in ('exit', 'exit()'):
                if memory_enable:
                    print('Encoding the our story...')
                    memory_ext.encode_memory()
                return

            # append chat to history.
            chat_history.append({"role": "user", "content": prompt,})

            kwargs = {}
            if memory_enable:
                memories: list = memory_ext.ask_query(prompt, top_k_memory)
                if memories:
                    kwargs = {'memories': memories}

            reply = self._chat(chat_history, text_generation_model,\
                tokenizer_model, max_new_tokens, **kwargs)

            chat_history.append(reply)

            temp_path: str = os.path.join\
                (self.utils_ext.user_path, user, 'history', f'{chat_name}.json')
            self.utils_ext.write_json_file(temp_path, chat_history)

    def docs_question_answer(
            self,
            text_generation_model_name: str = '',
            max_new_tokens: int = 500,
            sentence_transformer_model_name: str = '',
            cross_encoder_model_name: str = '',
            top_k: int = 0,
            allow_score: float = 0.0,
            encode_at_start: bool = False,
            show_retrieve: bool = False,
        ):
        """
        Ask information from provided docs.

        Args:
            text_generation_model_name (str): text generation model's name, use config if blank.
            max_new_tokens (int): max tokens to generate, default as 500.
            sentence_transformer_model_name (str): sentence transformer model's name, \
                use config if blank.
            cross_encoder_model_name (str): cross encoder model's name, use config if blank.
            top_k (int, optional): how many sentences you want to retrieve.
            allow_score (float, optional): retrieving process will stop when \
                similiarity score is lower.
            encode_at_start (bool, optional): encode memory before chating.
            show_retrieve (bool, optional): show retrieved data.
        """
        self.config.get_config_file()

        if text_generation_model_name == '':
            self.config.check_for_exist_model(1)
            text_generation_model_name = self.config.data['models']['Text_Generation']
            logging.info('No text generation model, use %s instead.', text_generation_model_name)

        if sentence_transformer_model_name == '':
            self.config.check_for_exist_model(2)
            sentence_transformer_model_name = self.config.data['models']['Sentence_Transformer']
            logging.info('No sentence transformer model, \
use %s instead.', sentence_transformer_model_name)

        if cross_encoder_model_name == '':
            self.config.check_for_exist_model(3)
            cross_encoder_model_name = self.config.data['models']['Cross_Encoder']
            logging.info('No cross encoder model, use %s instead.', cross_encoder_model_name)

        history: list = [
            {
                "role": "system", 
                "content": "You are an Assistant named LocalAssistant (Locas). \
Got provided with tons of docs, your duty is answering user's questions the best as possible. \
If docs' data are nonesense, you can ignore them and use your own words."
            },
        ]

        # load model
        logging.debug('Begin to load models.')
        text_generation_model, tokenizer_model = self._load_local_model(text_generation_model_name)
        docs_ext = DocsQuestionAnswerExtension\
                (sentence_transformer_model_name, cross_encoder_model_name)
        if encode_at_start:
            print("Encoding at start. Please be patient, it may take some minutes.")
            docs_ext.encode()
        logging.debug('Done loading models.')

        print(f"Start docs Q&A with '{text_generation_model_name}' for text generation, \
'{sentence_transformer_model_name}' for sentence transformer.\n\nType 'exit' to exit.", end='')

        while True:
            prompt: str = input('\n\n>> ')
            print()

            if prompt.lower() in ('exit', 'exit()'):
                return

            docs_data: list = docs_ext.ask_query(prompt, top_k, allow_score)
            docs_dict: dict = {}
            for data in docs_data:
                try:
                    docs_dict[data['title']].append(data['content'])
                except KeyError:
                    docs_dict.update({data['title']: [data['content']]})  

            prompt_input: str = "Retrieved data from docs:\n"
            for index, (title, content) in enumerate(docs_dict.items()):
                prompt_input += f"{index}. From file '{title}':\n"
                for text in content:
                    prompt_input += f"  - {text}\n"
                prompt_input += '\n'

            if show_retrieve:
                div: int = os.get_terminal_size().columns * '-'
                print(f'{div}\n{prompt_input}{div}\n')

            prompt_input += f"\nQuestion: {prompt}"
            history.append({"role": "user", "content": prompt_input,})

            reply = self._chat(history, text_generation_model, tokenizer_model, max_new_tokens)

            history.append(reply)



"""Used for retrive information from documents."""

import os
import logging
import pathlib
import shutil
import re

import pymupdf
import torch
from sentence_transformers import SentenceTransformer, CrossEncoder, util

from ..utils import ConfigManager, LocalAssistantException

class DocsQuestionAnswerExtension:
    """Extension used to process document."""
    def __init__(self, sentence_transformers_name: str, cross_encoder_name: str):
        self.config = ConfigManager()
        self.utils_ext = self.config.utils_ext

        try:
            temp_path: str = os.path.join\
                (self.utils_ext.model_path, 'Sentence_Transformer', sentence_transformers_name)
            self.sentence_transformers = SentenceTransformer(temp_path, local_files_only=True)

            temp_path: str = os.path.join\
                (self.utils_ext.model_path, 'Cross_Encoder', cross_encoder_name)
            self.cross_encoder = CrossEncoder(temp_path, local_files_only=True)
        except Exception as err:
            logging.error('Can not load model due to: %s', err)
            raise LocalAssistantException('Can not load model.') from err

    @staticmethod
    def _split_into_sentences(text: str) -> list[str]:
        """
        Split the text into sentences.

        If the text contains substrings "<prd>" or "<stop>", they would lead 
        to incorrect splitting because they are used as markers for splitting.
        
        Source: https://stackoverflow.com/questions/4576077/how-can-i-split-a-text-into-sentences
        """
        alphabets= "([A-Za-z])"
        prefixes = "(Mr|St|Mrs|Ms|Dr)[.]"
        suffixes = "(Inc|Ltd|Jr|Sr|Co)"
        starters = r"(Mr|Mrs|Ms|Dr|Prof|Capt|Cpt|Lt|He\s|She\s|It\s\
|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever)"
        acronyms = "([A-Z][.][A-Z][.](?:[A-Z][.])?)"
        websites = "[.](com|net|org|io|gov|edu|me)"
        digits = "([0-9])"
        multiple_dots = r'\.{2,}'

        text = " " + text + "  "
        text = text.replace("\n"," ")
        text = re.sub(prefixes,"\\1<prd>",text)
        text = re.sub(websites,"<prd>\\1",text)
        text = re.sub(f"{digits}[.]{digits}","\\1<prd>\\2",text)
        text = re.sub(multiple_dots, lambda match: "<prd>" * len(match.group(0)) + "<stop>", text)
        if "Ph.D" in text:
            text = text.replace("Ph.D.","Ph<prd>D<prd>")
        text = re.sub(fr"\s{alphabets}[.] "," \\1<prd> ",text)
        text = re.sub(f"{acronyms} {starters}","\\1<stop> \\2",text)
        text = re.sub(f"{alphabets}[.]{alphabets}[.]{alphabets}[.]","\\1<prd>\\2<prd>\\3<prd>",text)
        text = re.sub(f"{alphabets}[.]{alphabets}[.]","\\1<prd>\\2<prd>",text)
        text = re.sub(f" {suffixes}[.] {starters}"," \\1<stop> \\2",text)
        text = re.sub(f" {suffixes}[.]"," \\1<prd>",text)
        text = re.sub(f" {alphabets}[.]"," \\1<prd>",text)
        if "”" in text:
            text = text.replace(".”","”.")
        if "\"" in text:
            text = text.replace(".\"","\".")
        if "!" in text:
            text = text.replace("!\"","\"!")
        if "?" in text:
            text = text.replace("?\"","\"?")
        text = text.replace(".",".<stop>")
        text = text.replace("?","?<stop>")
        text = text.replace("!","!<stop>")
        text = text.replace("<prd>",".")
        sentences = text.split("<stop>")
        sentences = [s.strip() for s in sentences]
        if sentences and not sentences[-1]:
            sentences = sentences[:-1]
        return sentences

    def _get_file(self, path: str) -> list:
        """To get file, for `_get_docs()`."""
        path: pathlib.Path = pathlib.Path(path)

        result: list = []
        if not path.is_dir():
            return [path]
        for item in os.scandir(path):
            result += self._get_file(item.path)
        return result

    def _get_docs(self) -> tuple:
        """
        Get all documents that currently exist.

        Returns:
            tuple: list of exist docs (if is dir, go inside).
        """
        result: list = []

        logging.debug("Get path from 'share_path.json'.")
        try: # get all path that doesn't get copy.
            data: list = self.utils_ext.read_json_file\
                (os.path.join(self.utils_ext.docs_path, 'share_path.json'))
            for path in data:
                if pathlib.Path(path).exists():
                    result += self._get_file(path)
        except FileNotFoundError:
            pass

        logging.debug("Get path from copied documents.")
        # get those copied.
        for item in os.scandir(self.utils_ext.docs_path):
            if item.name in ('share_path.json', 'docs_data.json', 'encoded_docs_data.pt'):
                continue
            result += self._get_file(item.path)
        return tuple(result)

    def upload_docs(self, path: str, copy: bool = False) -> None:
        """
        Upload documents for LocalAssistant to use.

        Args:
            path (str): used to locate documents' path.
            copy (bool, optional): Whether to copy file/folder. Defaults to False.
        """
        print("Loading datasets. Please be patient, it may take some minutes.")

        path: pathlib.Path = pathlib.Path(path)
        if not path.exists():
            logging.error("'%s' is not a valid path.", path)
            raise LocalAssistantException(f"'{path}' is not a valid path.")

        try:
            os.makedirs(self.utils_ext.docs_path)
        except FileExistsError:
            pass

        if not copy: # we only need path.
            logging.debug("Put path in 'share_path.json' file.")
            try:
                total_path = self.utils_ext.read_json_file\
                    (os.path.join(self.utils_ext.docs_path, 'share_path.json'))
            except FileNotFoundError:
                total_path: list = []
            total_path.append(str(path))
            self.utils_ext.write_json_file\
                (os.path.join(self.utils_ext.docs_path, 'share_path.json'), total_path)
            self.encode()
            return

        # copy them to `documents` folder.
        logging.debug("Copy data to documents.")
        if path.is_dir():
            shutil.copytree\
                (path, os.path.join(self.utils_ext.docs_path, path.name), dirs_exist_ok=True)
        else:
            shutil.copyfile\
                (path, os.path.join(self.utils_ext.docs_path, path.name))
        self.encode()

    def encode(self) -> None:
        """Encode documents."""
        self.config.get_config_file()

        logging.info('Get existed docs.')
        docs: tuple = self._get_docs()

        # save memory to json file.
        logging.debug("Transfer data to .json")
        data_to_json: dict = {}
        index: int = -1
        data: list = []

        for doc in docs:
            if str(doc).lower().endswith\
                    (('.pdf','.xps', '.epub', '.mobi', '.fb2', '.cbz', '.svg', '.txt')):
                pdf_file = pymupdf.open(doc)
                docs_data = '\n'.join([page.get_text() for page in pdf_file])
            else:
                try:
                    file = open(doc, mode="r", encoding="utf-8")
                    docs_data = file.read()
                    file.close()
                except UnicodeDecodeError:
                    logging.error("Can not read file '%s'", pathlib.Path(doc).name)
                    continue
            # add data to json.
            for sentence in self._split_into_sentences(docs_data):
                if sentence in data: # prevent repetiton.
                    continue

                index += 1
                data.append(sentence)

                data_to_json.update({
                    index: {
                        "title": pathlib.Path(doc).name,
                        "content": sentence,
                    }
                })

        temp_path = os.path.join(self.utils_ext.docs_path, 'docs_data.json')
        self.utils_ext.write_json_file(temp_path, data_to_json)

        # encode
        encoded_data = self.sentence_transformers\
            .encode(data, convert_to_tensor=True, show_progress_bar=True)

        temp_path = os.path.join(self.utils_ext.docs_path, 'encoded_docs_data.pt')
        torch.save(encoded_data, temp_path)

    def ask_query(self, question: str, top_k: int = 0, allow_score: float = 0.0) -> list[str]:
        """Ask query function."""
        self.config.get_config_file()

        if top_k == 0: # user didnt add top k when chat.
            top_k = int(self.config.data["documents"]["top_k"])
        if allow_score == 0.0:
            allow_score = float(self.config.data["documents"]["allow_score"])

        temp_path: str = os.path.join(self.utils_ext.docs_path, 'encoded_docs_data.pt')
        logging.info('Loading docs from encoded data.')
        try:
            encoded_data = torch.load(temp_path, weights_only=True)
        except FileNotFoundError:
            logging.info('Encoded data not found. Encode new data.')
            self.encode()
            encoded_data = torch.load(temp_path, weights_only=True)

        # encode query.
        encoded_question = self.sentence_transformers.encode(question, convert_to_tensor=True)

        scores = util.dot_score(encoded_question, encoded_data)[0].tolist()
        hits = list(zip(range(len(scores)), scores))
        hits = sorted(hits, key=lambda x: x[1], reverse=True)

        # get json file.
        logging.debug('Get json file.')
        temp_path = os.path.join(self.utils_ext.docs_path, 'docs_data.json')
        data = self.utils_ext.read_json_file(temp_path)

        # Retrieve data using sentence transformers.
        retrieve_data: list = []
        for hit_data, hit_score in hits[:200]: # maximum 200 lines.
            if hit_score < allow_score:
                break
            retrieve_data.append((question, data[str(hit_data)]['content']))

        # Re-rank using cross encoder.
        scores = self.cross_encoder.predict(retrieve_data)

        hits = list(zip(hits, scores))
        hits = sorted(hits, key=lambda x: x[1], reverse=True)

        # Compile results.
        result: list = []
        for (hit_data, _), _ in hits[:top_k]:
            pointer: dict = data[str(hit_data)]
            result.append(pointer)
            logging.debug("Retrieve '%s' from '%s'.", pointer["content"], pointer["title"])

        return result



"""__init__ of the package."""

from .utils import UtilsExtension, ConfigManager, LocalAssistantException
from .model_processor import ChatExtension, DownloadExtension, MemoryExtension

# check for PyTorch
try:
    import torch
except ImportError as err:
    raise LocalAssistantException\
        ("Could not find torch installed. Please visit https://pytorch.org/ \
and download the version for your device.") from err

__all__ = [
    'UtilsExtension',
    'ChatExtension',
    'DownloadExtension',
    'MemoryExtension',
    'ConfigManager',
    'LocalAssistantException',
]

__version__ = '1.1.0rc1'



"""Main process of LocalAssistant"""

import logging
import argparse
import os
import shutil

from .parser import PARSER
from .utils import LocalAssistantException
from .model_processor import DownloadExtension, ChatExtension, DocsQuestionAnswerExtension

chat_ext = ChatExtension()
utils_ext = chat_ext.utils_ext
config = chat_ext.config

config.get_config_file()

def _download(parser_arg: argparse.Namespace):
    """download command function"""

    if parser_arg.TASK == 0:
        return

    try:
        task: int = int(parser_arg.TASK)
    except ValueError: # is string
        try:
            task: int = utils_ext.reverse_model_task[parser_arg.TASK]
        except KeyError as err: # not found
            logging.error("Invalid TASK: '%s'.", parser_arg.TASK)
            raise LocalAssistantException(f"invalid TASK: '{parser_arg.TASK}'.") from err
    if not 0 < task <= len(utils_ext.model_task):
        logging.error("Invalid TASK: '%s'.", parser_arg.TASK)
        raise LocalAssistantException(f"invalid TASK: '{parser_arg.TASK}'.")

    DownloadExtension().download_model_from_huggingface\
        (parser_arg.name, parser_arg.PATH, parser_arg.token, task)

def _config(parser_arg: argparse.Namespace):
    """Config command function"""

    def _change_single(name: str, name_help: str, condition = None) -> bool:
        """
        Fast VALUE changer.
        
        Returns:
            bool: whether function executed successfully.
        """
        print(f"{name_help}\n\nModify VALUE of '{name}' to ... (Type 'exit' to exit.)\n")
        command: str = input('>> ')
        print()

        # for exit, not everyone remember their token anyway.
        if command.lower() in ('exit', 'exit()'):
            return False

        if condition is not None:
            if not condition(command):
                logging.error("Invalid VALUE: '%s'.", command)
                raise LocalAssistantException(f"Invalid VALUE: '{command}'.")

        config.data.update({name: command})
        config.upload_config_file()
        return True

    # show config data.
    if parser_arg.show:
        config.print_config_data()
        return

    # modify config data.
    command: str = ''
    while True:
        config.print_config_data()
        print("Type KEY to modify KEY's VALUE. Type 'exit' to exit.\n")
        command = input('>> ')
        command = command.lower()
        print()

        if command in ('exit', 'exit()'):
            break

        if command not in tuple(config.data.keys()):
            logging.error("Invalid KEY: '%s'.", command)
            raise LocalAssistantException(f"Invalid KEY: '{command}'.")

        match command:
            case 'hf_token':
                if not _change_single(command, "'hf_token' is your Hugging Face token. \
Some models might be restricted and need authenticated."):
                    break
                continue

            case 'load_in_bits':
                def _check_valid(command: str) -> bool:
                    if command in ('4', '8', 'None'):
                        return True
                    return False

                if not _change_single(command, "'load_in_bits' is for 'quantization' method. \
If the VALUE is 8, then model is load in 8 bits (1 bytes) per parameters. \
Choose from: '4', '8', 'None'.", _check_valid):
                    break
                continue

            case 'top_k_memory':
                def _check_valid(command: str) -> bool:
                    if int(command) > 0:
                        return True
                    return False

                if not _change_single(command,\
                    "'top_k_memory' let us know how much memory you want to recall.", _check_valid):
                    break
                continue

            case 'models':
                while True:
                    utils_ext.print_dict(config.data['models'])
                    print("\nType KEY to modify KEY's VALUE. Type 'exit' to exit.\n")
                    command = input('>> ')
                    print()

                    if command.lower() in ('exit', 'exit()'):
                        break

                    if command not in tuple(config.data['models'].keys()):
                        logging.error("Invalid KEY: '%s'.", command)
                        raise LocalAssistantException(f"Invalid KEY: '{command}'.")

                    while True:
                        print('Choose from:')

                        model: str = config.data['models'][command]
                        folder_model: list = []
                        for item in os.scandir(os.path.join(utils_ext.model_path, model)):
                            if item.is_dir():
                                folder_model.append(item.name)
                                print(f'    - {item.name}')

                        if not _change_single(model, ''):
                            break
                        continue

            case 'documents':
                while True:
                    utils_ext.print_dict(config.data['documents'])
                    print("\nType KEY to modify KEY's VALUE. Type 'exit' to exit.\n")
                    command = input('>> ')
                    print()

                    if command.lower() in ('exit', 'exit()'):
                        break

                    if command not in tuple(config.data['documents'].keys()):
                        logging.error("Invalid KEY: '%s'.", command)
                        raise LocalAssistantException(f"Invalid KEY: '{command}'.")

                    if command == 'top_k':
                        def _check_valid(command: str) -> bool:
                            if 0 < int(command) < 51:
                                return True
                            return False

                        if not _change_single(command,"'top_k' let us know how many \
lines you want to retrieve. Maximum is 50 lines.", _check_valid):
                            break
                        continue

                    if command == 'allow_score':
                        def _check_valid(command: str) -> bool:
                            if 0 <= float(command) <= 1:
                                return True
                            return False

                        if not _change_single(command,"'allow_score' will make the retrieving \
process stop when similiarity score is lower.", _check_valid):
                            break
                        continue

            case 'users':
                print("Type 'locas user -h' for better configuration.\n")
                continue

def _user(parser_arg: argparse.Namespace):
    """User command function."""

    exist = config.check_exist_user_physically(parser_arg.TARGET)

    # show user.
    if parser_arg.TARGET.lower() == 'show':
        users: list = []
        for item in os.scandir(utils_ext.user_path):
            if not item.is_dir():
                continue
            if config.check_exist_user_physically(item.name):
                users.append(item.name)
        if users:
            print('Existed users:')
        for user in users:
            print(f'    - {user}')

    # create user.
    elif parser_arg.create:
        if exist:
            logging.error("User '%s' is existed.", parser_arg.TARGET)
            raise LocalAssistantException(f"User '{parser_arg.TARGET}' is existed.")
        config.check_for_exist_user(parser_arg.TARGET)
        logging.info('Created user %s.', parser_arg.TARGET)

    # delete user.
    elif parser_arg.delete:
        if not exist:
            logging.error("User '%s' is not existed.", parser_arg.TARGET)
            raise LocalAssistantException(f"User '{parser_arg.TARGET}' is not existed.")
        shutil.rmtree(os.path.join(utils_ext.user_path, parser_arg.TARGET))
        logging.info('Deleted user %s.', parser_arg.TARGET)

        # rename user.
    elif parser_arg.rename is not None:
        if not exist:
            logging.error("User '%s' is not existed.", parser_arg.TARGET)
            raise LocalAssistantException(f"User '{parser_arg.TARGET}' is not existed.")

        if config.check_exist_user_physically(parser_arg.rename):
            logging.error("User '%s' is existed.", parser_arg.rename)
            raise LocalAssistantException(f"User '{parser_arg.rename}' is existed.")

        os.rename(os.path.join(utils_ext.user_path, parser_arg.TARGET),\
                  os.path.join(utils_ext.user_path, parser_arg.rename))
        logging.info('Renamed user %s to %s.', parser_arg.TARGET, parser_arg.rename)

    # change user.
    else:
        if not exist:
            logging.error("User '%s' is not existed.", parser_arg.TARGET)
            raise LocalAssistantException(f"User '{parser_arg.TARGET}' is not existed.")
        config.data.update({'users': parser_arg.TARGET})
        logging.info('Change user to %s.', parser_arg.TARGET)

def _chat(parser_arg: argparse.Namespace):
    """Chat command function."""

    if parser_arg.LINE < 1:
        logging.error("Invalid LINE: %s", parser_arg.LINE)
        raise LocalAssistantException(f"Invalid LINE: {parser_arg.LINE}")

    chat_ext.chat_with_limited_lines\
        (parser_arg.text_generation, parser_arg.LINE, parser_arg.max_token)

def _start(parser_arg: argparse.Namespace):
    """Start command function."""

    chat_ext.chat_with_history(parser_arg.text_generation, parser_arg.user, parser_arg.max_token,\
        parser_arg.memory_enable, parser_arg.sentence_transformer,\
        parser_arg.top_k_memory, parser_arg.encode_at_start)

def _docs(parser_arg: argparse.Namespace):
    """Docsqa command function."""

    if parser_arg.ACTION == 'upload':
        config.check_for_exist_model(2)
        DocsQuestionAnswerExtension\
            (config.data['models']['Sentence_Transformer'],config.data['models']['Cross_Encoder'])\
            .upload_docs(parser_arg.PATH, parser_arg.copy)

    else:
        chat_ext.docs_question_answer(parser_arg.text_generation, parser_arg.max_token,\
            parser_arg.sentence_transformer, parser_arg.cross_encoder, parser_arg.top_k,\
            parser_arg.allow_score, parser_arg.encode_at_start, parser_arg.show_retrieve)

def _self_destruction(parser_arg: argparse.Namespace):
    """Self-destruction function."""

    utils_ext.self_destruction(parser_arg.all)

def main():
    """Main, the one and only."""
    parser_arg: argparse.Namespace = PARSER.parse_args()

    verbose = 4 if parser_arg.verbose > 4 else parser_arg.verbose # limit to 4
    logging.basicConfig(level=(5-verbose) * 10, format="%(asctime)s [%(levelname)s]: %(message)s")

    match parser_arg.COMMAND:
        case 'download':
            _download(parser_arg)
        case 'config':
            _config(parser_arg)
        case 'user':
            _user(parser_arg)
        case 'chat':
            _chat(parser_arg)
        case 'start':
            _start(parser_arg)
        case 'docs':
            _docs(parser_arg)
        case 'self-destruction':
            _self_destruction(parser_arg)

if __name__ == '__main__':
    main()



"""__init__ of model processor."""

from .chat import ChatExtension
from .download import DownloadExtension
from .memory import MemoryExtension
from .docs import DocsQuestionAnswerExtension

__all__ = [
    'ChatExtension',
    'DownloadExtension',
    'MemoryExtension',
    'DocsQuestionAnswerExtension',
]


"""Parser, define when goes `locas -h`."""

import argparse

PARSER = argparse.ArgumentParser(
    prog='locas',
    description='LocalAssistant (locas) is an AI designed to be used in CLI.',
)

# verbose.
PARSER.add_argument('-v', '--verbose', action='count', help='show debug \
messages (Can be used multiple times for higher level: CRITICAL[v] -> DEBUG[vvvv])', default=0)

# version.
PARSER.add_argument('-V', '--version', action='version', version='LocalAssistant 1.1.0rc1')

subparser = PARSER.add_subparsers(
    title='commands',
    description="built-in commands (type 'locas COMMAND -h' for better description).",
    metavar='COMMAND',
    dest='COMMAND',
)

# +--------------------+
# | locas download ... |
# +--------------------+

subparser_download = subparser.add_parser(
    name='download',
    help='Download models from Hugging Face',
    description='Download models from Hugging Face',
    formatter_class=argparse.RawTextHelpFormatter,
)

subparser_download.add_argument('PATH', action='store', help='Path of the Hugging Face\'s model')

TEMP_STRING: str = """\
Model\'s task. Choose from:
    - 'Text_Generation' (or '1'): Download text generation model.
    - 'Sentence_Transformer' (or '2'): Download sentence transformer model.
    - 'Cross_Encoder' (or '3'): Download cross encoder model.
"""
subparser_download.add_argument('TASK', action='store', help=TEMP_STRING, default=0)
del TEMP_STRING

subparser_download.add_argument('-n', '--name', action='store',\
    help='Name of the model to be saved', default='Untitled')

subparser_download.add_argument('-t', '--token', action='store', help='User Hugging \
Face\'s token (Some models might be restricted and need authenticated)', default='')

# +------------------+
# | locas config ... |
# +------------------+

TEMP_STRING: str = """\
Configurate LocalAssistant.
 
----------------------------Example-----------------------------------
>> locas config -m

"hf_token": "",
"load_in_bits": "8",
"top_k_memory": "5",
"models": {
    "Text_Generation": "",
    "Sentence_Transformer": "",
    "Cross_Encoder": "",
},
"documents": {
    "top_k": "10",
    "allow_score": "0.8",
},
"users": "default"

Type KEY to modify KEY's VALUE. Type 'exit' to exit.

>> load_in_bits

'load_in_bits' is for 'quantization' method. If the VALUE is 8, then model is load in 8 bits (1 bytes) per parameters. Choose from: '4', '8', 'None'.

Modify VALUE of 'load_in_bits' to ... (Type 'exit' to exit.)

>> None
----------------------------------------------------------------------

"""
subparser_config = subparser.add_parser(
    name='config',
    help='Configurate LocalAssistant.',
    description=TEMP_STRING,
    formatter_class=argparse.RawTextHelpFormatter,
)
del TEMP_STRING

subparser_config_group = subparser_config.add_mutually_exclusive_group(required=True)

subparser_config_group.add_argument('-m', '--modify', action='store_true',\
    help='Modify config value')

subparser_config_group.add_argument('-s', '--show', action='store_true', help='Show config data')

# +----------------+
# | locas user ... |
# +----------------+

TEMP_STRING: str = """\
Use this to configurate user.
    - To change change user. Type 'locas user TARGET'.
    - To do other stuff, use (-c|-d|-r NAME).
    - To show exist user. Type 'locas user show'.
"""
subparser_user = subparser.add_parser(
    name='user',
    help='Config user.',
    description=TEMP_STRING,
    formatter_class=argparse.RawDescriptionHelpFormatter
)
del TEMP_STRING

subparser_user.add_argument('TARGET', action='store', help='The target')

subparser_user_group = subparser_user.add_mutually_exclusive_group()

subparser_user_group.add_argument('-c', '--create', action='store_true',\
    help='Create user with TARGET name')

subparser_user_group.add_argument('-d', '--delete', action='store_true',\
    help='Delete user with TARGET name')

subparser_user_group.add_argument('-r', '--rename', action='store', metavar='NAME',\
    help='Rename TARGET with NAME')

# +----------------+
# | locas chat ... |
# +----------------+

subparser_chat = subparser.add_parser(
    name='chat',
    help='Chat with models for limited lines. (no history saved)',
    description='Chat with models for limited lines. \
Recommend for fast chat as non-user. (no history saved)',
)

subparser_chat.add_argument('LINE', action='store', type=int, help='Number of line to chat with')

subparser_chat.add_argument('-tgm', '--text-generation', metavar='MODEL', action='store',\
    help='Use downloaded text generation model', default='')

subparser_chat.add_argument('-t', '--max-token', metavar='TOKEN', action='store', type=int,\
    help='Max tokens to generate', default= 500)

# +-----------------+
# | locas start ... |
# +-----------------+

subparser_start = subparser.add_parser(
    name='start',
    help='Chat with models using history.',
    description='Chat with models using history.',
)

subparser_start.add_argument('-u', '--user', action='store',\
    help='The user name', default='default')

subparser_start.add_argument('-tgm', '--text-generation', metavar='MODEL', action='store',\
    help='Use downloaded text generation model', default='')

subparser_start.add_argument('-t', '--max-token', metavar='TOKEN', action='store', type=int,\
    help='Max tokens to generate', default= 500)

subparser_start.add_argument('-m', '--memory-enable', action='store_true',\
    help='Enable memory function')

subparser_start.add_argument('-stm', '--sentence-transformer', metavar='MODEL', action='store',\
    help='Use downloaded sentence transformer model. (When memory enabled)', default='')

subparser_start.add_argument('-tk', '--top-k-memory', metavar='TOP_K', action='store', type=int,\
    help='How much memory you want to recall. (When memory enabled)', default= 0)

subparser_start.add_argument('--encode-at-start', action='store_true',\
    help='Encode memory before chating. (When memory enabled)')

# +----------------+
# | locas docs ... |
# +----------------+

subparser_docs = subparser.add_parser(
    name='docs',
    help='Ask information from provided documents.',
    description='Ask information from provided documents.',
)

subparser_docs = subparser_docs.add_subparsers(
    title='actions',
    description='Action to do with documents.',
    metavar='ACTION',
    dest='ACTION',
)

subparser_docs_upload = subparser_docs.add_parser(
    name='upload',
    help='Upload files/folders to documents.',
    description='Upload files/folders to documents.',
)

subparser_docs_upload.add_argument('PATH', action='store', help='Path to add.')

subparser_docs_upload.add_argument('-c', '--copy', action='store_true',\
    help='Copy provided folders/files to docs.')

subparser_docs_chat = subparser_docs.add_parser(
    name='chat',
    help='Ask queries from docs and get answer.',
    description='Ask queries from docs and get answer.'
)

subparser_docs_chat.add_argument('-tgm', '--text-generation', metavar='MODEL', action='store',\
    help='Use downloaded text generation model', default='')

subparser_docs_chat.add_argument('-t', '--max-token', metavar='TOKEN', action='store', type=int,\
    help='Max tokens to generate', default= 500)

subparser_docs_chat.add_argument('-stm', '--sentence-transformer',metavar='MODEL',action='store',\
    help='Use downloaded sentence transformer model.', default='')

subparser_docs_chat.add_argument('-cem', '--cross-encoder',metavar='MODEL',action='store',\
    help='Use downloaded cross encoder model.', default='')

subparser_docs_chat.add_argument('-tk', '--top-k', metavar='TOP_K', action='store', type=int,\
    help='How many sentences you want to retrieve.', default= 0)

subparser_docs_chat.add_argument('-s', '--allow-score',metavar='SCORE',action='store',type=float,\
    help='Retrieving process will stop when similiarity score is lower.', default= 0.0)

subparser_docs_chat.add_argument('--encode-at-start', action='store_true',\
    help='Encode docs before chating.')

subparser_docs_chat.add_argument('--show-retrieve', action='store_true',\
    help='Show the retrieved data from docs.')

# +----------------------------+
# | locas self-destruction ... |
# +----------------------------+

subparser_self_destruction = subparser.add_parser(
    name='self-destruction',
    help='LocalAssistant\'s self-destruction.',
    description='LocalAssistant\'s self-destruction.',
)

subparser_self_destruction.add_argument('-a', '--all', action='store_true',\
    help='Delete the whole folder (included models, history, etc).')



"""Help downloading models."""

import logging
import os
import pathlib

from transformers import AutoTokenizer, AutoModelForCausalLM
from sentence_transformers import SentenceTransformer, CrossEncoder

from ..utils import UtilsExtension, LocalAssistantException

class DownloadExtension:
    """Download extention for LocalAssistant."""
    def __init__(self):
        self.utils_ext = UtilsExtension()

    @staticmethod
    def _download_model(
            huggingface_path: str,
            model,
            hf_token: str = '',
        ):
        """
        Some models might be restricted and need authenticated. \
            Use token to login temporately and download model.

        Args:
            huggingface_path (str): path to huggingface model (author/model_name)
            model (Any): model is used.
            hf_token (str): huggingface's token.
        """
        if hf_token == '': # by default, do not use token.
            hf_token: None = None
            logging.debug('Not use token.')
        else:
            logging.debug('Use provided token: %s.', hf_token)

        try:
            if isinstance(model, SentenceTransformer):
                return SentenceTransformer(huggingface_path, token=hf_token)
            if model == 'CrossEncoder':
                return CrossEncoder(huggingface_path)
            return model.from_pretrained\
                (huggingface_path, use_safetensors=True, device_map="auto", token=hf_token)
        except Exception as err:
            logging.error('Can not download model due to: %s.', err)
            raise LocalAssistantException('Can not download model.') from err


    @staticmethod
    def _save_model(
            model:AutoTokenizer|AutoModelForCausalLM|SentenceTransformer,
            path: str|pathlib.Path,
        ) -> str:
        """
        Save model to path. Check if the name has taken and rename. (only dir)

        Args:
            model (AutoTokenizer | AutoModelForCausalLM | SentenceTransformer): model is used.
            path (str | pathlib.Path): path to save to.

        Returns:
            str: path after fixed.
        """
        # take parent and child path
        path = pathlib.Path(path)

        parent: pathlib.Path = path.parent
        child: str = path.name

        try: # make dir if dir not exist
            os.makedirs(parent)
            logging.debug('Made %s directory.', parent.name)
        except FileExistsError:
            pass

        stop: bool = False
        scanned: bool = False
        while not stop:
            for item in os.scandir(parent):
                scanned = True

                if item.name == child:
                    if not item.is_dir():
                        continue
                    logging.debug('Found %s.', child)
                    index: str = item.name.split(' ')[-1]

                    # check if index in (n) format
                    if not index.startswith('(') or not index.endswith(')'):
                        child += ' (1)'
                        break

                    try:
                        index: int = int(index[1:-1])
                    except ValueError: # it was (n) but n is not int
                        child += ' (1)'
                        break

                    child = f'{child[:-4]} ({index + 1})'
                    break
                else:
                    stop = True
            if not scanned: # not scanned mean dir is empty.
                break
        del stop, scanned

        logging.info('Save as %s in %s.', child, parent.name)
        model.save_pretrained(str(parent / child))
        return str(parent / child)

    def download_model_from_huggingface(
            self,
            model_name: str,
            huggingface_path: str,
            hf_token: str = '',
            task: int = 0,
        ) -> None:
        """
        Download model directly from Hugging Face and save it in `models` folder.

        Args:
            model_name (str): The name of models. Used for select model and other config.
            huggingface_path (str): The path to download model.
            hf_token (str): The user Hugging Face access token.
            task (int): Model's task. Defined in `utils_ext.model_task`.
        """
        # if there is no task, return.
        if task == 0:
            return

        # Download model from huggingface path.

        # if user use 'https' path, convert to normal one.
        huggingface_path = huggingface_path.removeprefix('https://huggingface.co/')

        cur_task: str = self.utils_ext.model_task[task].lower().replace('_', ' ')
        logging.info('Download %s model from %s.',cur_task, huggingface_path)
        del cur_task

        match task:
            case 1: # For text generation.
                tokenizer_model = self._download_model(huggingface_path, AutoTokenizer, hf_token)
                text_generation_model = self\
                    ._download_model(huggingface_path, AutoModelForCausalLM, hf_token)

                # save downloaded model
                downloaded_path: str = os.path.join\
                    (self.utils_ext.model_path, 'Text_Generation', model_name)
                downloaded_path = self._save_model(text_generation_model, downloaded_path)
                self._save_model(tokenizer_model, os.path.join(downloaded_path, 'Tokenizer'))

            case 2: # For sentence transformer
                sentence_transformer_model = self\
                    ._download_model(huggingface_path, SentenceTransformer(), hf_token)

                # save downloaded model
                downloaded_path: str = os.path.join\
                    (self.utils_ext.model_path, 'Sentence_Transformer', model_name)
                self._save_model(sentence_transformer_model, downloaded_path)

            case 3: # For cross encoder
                cross_encoder_model = self._download_model\
                    (huggingface_path, 'CrossEncoder', hf_token) # CrossEncoder require input.

                # save downloaded model
                downloaded_path: str = os.path.join\
                    (self.utils_ext.model_path, 'Cross_Encoder', model_name)
                self._save_model(cross_encoder_model, downloaded_path)



"""Setup file."""

import pathlib
import setuptools

setuptools.setup(
    name="LocalAssistant",
    version="1.1.0rc1",
    description="LocalAssistant (locas) is an AI designed to be used in CLI. \
(Currently in development)",
    long_description=pathlib.Path('README.md').read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    author="Linos",
    project_urls={
        "Source": "https://github.com/Linos1391/LocalAssistant",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Environment :: Console",
        "Environment :: GPU :: NVIDIA CUDA",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
    ],
    python_requires=">=3.10",
    install_requires=[
        "transformers[torch]",
        "bitsandbytes",
        "sentence-transformers",
        "PyMuPDF",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={"console_scripts": ["locas = LocalAssistant.main:main"]},
)



"""Utils file."""

import pathlib
import json
import os
import stat
import logging

###### - Add more model:
# 1.n.n
# - SENTENCE TRANSFORMATION \ CROSS_ENCODER (Extract data from database)
# - Search google https://github.com/Nv7-GitHub/googlesearch
# 2.n.n
# - TEXT_TO_SPEECH (Voice from AI)
# - SPEECH_TO_TEXT (Voice chat)
# - AUDIO_CLASSIFICATION (Multiple user voice chat
#      + https://huggingface.co/speechbrain/spkrec-ecapa-voxceleb) I will cook.

class LocalAssistantException(Exception):
    """For common errors in LocalAssistant"""

class UtilsExtension:
    """Main utils extension of LocalAssistant"""
    def __init__(self):
        self.model_task: dict = {
            1: 'Text_Generation',
            2: 'Sentence_Transformer',
            3: 'Cross_Encoder',
        }
        self.reverse_model_task: dict = {k: v for v, k in self.model_task.items()}

        # path and stuffs
        self.project_path: str = pathlib.Path(__file__).parent

        self.env_path: pathlib.Path = self.project_path
        while self.env_path.name != '.venv': # goes back until reach .venv
            self.env_path = self.env_path.parent

        self.model_path: str = self.env_path.parent / 'models'
        self.user_path: str = self.env_path.parent / 'users'
        self.docs_path: str = self.env_path.parent / 'documents'

    @staticmethod
    def _real_remove(path: str):
        """Something like shutil.rmtree but without access denied"""
        for root, dirs, files in os.walk(path, topdown=False):
            for name in files:
                filename = os.path.join(root, name)
                os.chmod(filename, stat.S_IWUSR)
                os.remove(filename)
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(path)

    @staticmethod
    def read_json_file(path: str):
        """
        Read a json file.

        Args:
            path (str): file's path.

        Returns:
            Any: file's data.
        """
        with open(path, mode="r", encoding="utf-8") as read_file:
            data = json.load(read_file)
            read_file.close()
        return data

    @staticmethod
    def write_json_file(path: str, data: dict) -> None:
        """
        Write a json file.

        Args:
            path (str): file's path.
            data (dict): file's data.
        """
        with open(path, mode="w", encoding="utf-8") as write_file:
            json.dump(data, write_file, indent=4)
            write_file.close()

    # it happens that i'm too smart.
    def print_dict(self, data: dict, level: int=0):
        """Used to print `dict` type."""
        for key, value in data.items():
            if isinstance(value, str):
                print(level*'   ' + f"'{key}': '{value}',")
            else: # value is dict
                print(level*'   ' + "%r: {" % (key)) # use %r as '{' is written
                self.print_dict(value, level+1)
                print(level*'   ' + "},")

    # remove all
    def self_destruction(self, delete_all: bool = False):
        """Everything all needs self-destruction."""
        option: str = input("Are you sure to remove LocalAssistant. \
There will be no turning back. Continue? [y/(n)]: ")
        if option.lower() != 'y':
            print('Self-destruction denied.')
            return
        print('Self-destruction...')

        # Locas, kys.

        if delete_all:
            self.env_path = self.env_path.parent
        self._real_remove(self.env_path)

    def load_chat_history(self, user: str) -> tuple[list, str]:
        """ 
        Load chat history!

        Args:
            user (str): user's name.
        
        Returns:
            tuple[list, str]: (list of history, history's name).
        """
        while True:
            scanned: bool = False
            history_list: list = []

            for _, _, files in os.walk(os.path.join(self.user_path, user, 'history')):
                if scanned:
                    break
                scanned = True

                if files == []:
                    print("\nThere is no history yet, please create one.")
                else:
                    print("\nChoose from:")
                for history in files:
                    if history.endswith('.json'):
                        print(f'    - {history.removesuffix('.json')}')
                        history_list.append(history.removesuffix('.json'))
            print("""
Type 'create [name (Required, 1 WORD ONLY] [system_prompt (Optional)]' to create new history.
Type 'delete [name (Required, 1 WORD ONLY]' to delete history.
Type 'exit' to exit.\n""")
            command: str = input('>> ')

            if command == '':
                logging.error('There is no input.')
                continue

            if command.lower() in ('exit', 'exit()'):
                break

            if command.split()[0].lower() == 'create':
                try:
                    command_split = command.split()
                    chat_name = command_split[1]
                    system_prompt = ' '.join(command_split[2:])
                except (ValueError, IndexError):
                    try:
                        chat_name = command.split()[1]
                    except IndexError:
                        logging.error('Missing NAME.')
                        continue

                if user == 'default':
                    system_prompt = "You are an Assistant named LocalAssistant (Locas). \
Give the user the best supports as you can."
                else:
                    system_prompt = f"You are an Assistant named LocalAssistant (Locas) \
who serves the user called {user}. Give {user} the best supports as you can."

                if chat_name in history_list: # throw error if create same name.
                    logging.error("ERROR: Name %s is used.\n", chat_name)
                    continue

                return ([{"role": "system", "content": system_prompt}], chat_name)

            if command.split()[0].lower() == 'delete':
                try:
                    chat_name = command.split()[1]
                except IndexError:
                    logging.error('Missing NAME.')
                    continue

                if chat_name not in history_list: # throw error if create same name.
                    logging.error('Name %s is not existed.', chat_name)
                    continue

                os.remove(os.path.join(self.user_path,user,'history',f'{chat_name}.json'))
                print()
                continue

            if command not in history_list:
                logging.error('No history named %s', command)
                continue

            print('\n')
            temp_path: str = os.path.join\
                (self.user_path, user, 'history', f'{command}.json')
            return ([v for v in self.read_json_file(temp_path)], command.split()[0])
        return ([], '')

class ConfigManager:
    """Config of LocalAssistant"""
    def __init__(self):
        self.utils_ext = UtilsExtension()
        self.data: dict = {}
        self.path: str = os.path.join(self.utils_ext.project_path, 'locas_config.json')

    def upload_config_file(self) -> None:
        """Dump data to `locas_config.json` file."""
        self.utils_ext.write_json_file(self.path, self.data)
        logging.debug("Uploaded current data to config.json file.")

    def get_config_file(self) -> None:
        """"Read data from `locas_config.json` file."""
        logging.debug('Finding locas_config.json file.')
        try:
            # Read the data
            self.data = self.utils_ext.read_json_file(self.path)
            logging.debug('Found locas_config.json file.')

        except FileNotFoundError:
            logging.debug('Cannot find locas_config.json file. Create new one.')

            self.data = {
                "hf_token": "", # Hugging Face token.
                "load_in_bits": "8", # 'quantization' method. (So the device won't blow up)
                "top_k_memory": "5", # num of memory to use
                "models": { # the model that being use for chatting.
                    "Text_Generation": "",
                    "Sentence_Transformer": "",
                    "Cross_Encoder": "",
                },
                "documents": {
                    "top_k": "10",
                    "allow_score": "0.6"
                },
                "users": "default",
            }

            # dump data to file.
            self.utils_ext.write_json_file(self.path, self.data)
        logging.debug('Got data from config file.')

    def print_config_data(self) -> None:
        """Print out current config data (not from file)."""
        self.utils_ext.print_dict(self.data)

    def check_exist_user_physically(self, target: str) -> bool:
        """
        Check user in folder.

        Args:
            target (str): user's name.

        Returns:
            bool: if user already existed.
        """
        scanned: bool = False
        for _, folders, _ in os.walk(os.path.join(self.utils_ext.user_path, target)):
            if scanned:
                break
            scanned = True

            if 'history' in folders and 'memory' in folders:
                return True
            return False

    def check_for_exist_user(self, user: str):
        """
        Check user in but in current data.

        Args:
            user (str): user's name.
        """
        if user == 'default': # user did not add 'user'.
            try:
                os.makedirs(os.path.join(self.utils_ext.user_path, 'default', 'history'))
            except FileExistsError:
                pass

            try:
                os.mkdir(os.path.join(self.utils_ext.user_path, 'default', 'memory'))
            except FileExistsError:
                pass

            self.data.update({'users': 'default',})
        else: # user add new user.
            self.data.update({'users': user})

            if not self.check_exist_user_physically(user):
                # update on physical directory
                try:
                    os.makedirs(os.path.join(self.utils_ext.user_path, user, 'history'))
                except FileExistsError:
                    pass
                os.mkdir(os.path.join(self.utils_ext.user_path, user, 'memory'))
            logging.debug("Created user '%s'.", user)
        self.upload_config_file()

    # check exist model for each task
    def check_for_exist_model(self, task: int) -> None:
        """Check for exist model."""
        if not 0 < task <= len(self.utils_ext.model_task):
            logging.error("Wrong task. Got '%s' (%d)", self.utils_ext.model_task[task], task)
            raise LocalAssistantException\
                (f"Wrong task. Got '{self.utils_ext.model_task[task]}' ({task})")

        task: str = self.utils_ext.model_task[task]

        if self.data['models'][task] != '':
            for root, folders, _ in os.walk(os.path.join(self.utils_ext.model_path, task)):
                if root != os.path.join(self.utils_ext.model_path, task):
                    break

                if self.data['models'][task] in folders:
                    return # nothing to fix.

        scanned: bool = False
        for _, folders, _ in os.walk(os.path.join(self.utils_ext.model_path, task)):
            if scanned:
                break

            scanned = True
            self.data['models'][task] = folders[0] # if no model, meaning folders == [] -> ignored.

        if not scanned: # the above line has skipped
            logging.critical("There is no models for %s. \
Please type 'locas download -h' and download one.", task)
            raise LocalAssistantException(f"There is no models for {task}. \
Please type 'locas download -h' and download one.")

        self.upload_config_file()
        logging.info('Apply %s as model for %s.',self.data['models'][task], task)
